/*
 * Author: Pavel Kirienko <pavel.kirienko@zubax.com>
 */

#pragma once

#include <os.hpp>
#include <cstdint>
#include <cassert>
#include <cstring>
#include <optional>

namespace bootloader_interface
{
/**
 * The CRC algorithm used by the bootloader to sign and verify shared data structures.
 *
 * CRC-64-WE
 * Description: http://reveng.sourceforge.net/crc-catalogue/17plus.htm#crc.cat-bits.64
 * Initial value: 0xFFFFFFFFFFFFFFFF
 * Poly: 0x42F0E1EBA9EA3693
 * Reverse: no
 * Output xor: 0xFFFFFFFFFFFFFFFF
 * Check: 0x62EC59E3F1A4F00A
 */
class CRC64WE
{
    static constexpr std::uint64_t Poly = 0x42F0E1EBA9EA3693ULL;
    static constexpr std::uint64_t Mask = std::uint64_t(1) << 63;

    std::uint64_t crc_ = 0xFFFFFFFFFFFFFFFFULL;

public:
    void add(const void* data, std::size_t len)
    {
        auto bytes = static_cast<const std::uint8_t*>(data);
        assert(bytes != nullptr);
        while (len-- > 0)
        {
            crc_ ^= std::uint64_t(*bytes++) << 56;
            for (std::uint8_t i = 0; i < 8; i++)
            {
                crc_ = (crc_ & Mask) ? (crc_ << 1) ^ Poly : crc_ << 1;
            }
        }
    }

    std::uint64_t get() const { return crc_ ^ 0xFFFFFFFFFFFFFFFFULL; }
};

/**
 * This class was taken out of the bootloader library of Zubax ChibiOS, and then it was greatly simplified.
 */
template <typename Container>
class AppSharedMarshaller
{
    void* location_;

    class ContainerWrapper
    {
        std::uint64_t crc_{};

    public:
        Container container{};

        ContainerWrapper() = default;

        ContainerWrapper(const Container& c) : container(c)
        {
            CRC64WE crc_computer;
            crc_computer.add(&container, sizeof(container));
            crc_ = crc_computer.get();
        }

        bool isValid() const
        {
            CRC64WE crc_computer;
            crc_computer.add(&container, sizeof(container));
            return crc_computer.get() == crc_;
        }
    };

public:
    AppSharedMarshaller(void* loc) : location_(loc) {}

    std::optional<Container> readAndErase()
    {
        ContainerWrapper wrapper;
        std::memmove(&wrapper, location_, sizeof(wrapper));
        if (wrapper.isValid())
        {
            std::memset(location_, 0, sizeof(wrapper));  // Erase
            return wrapper.container;
        }
        else
        {
            return {};
        }
    }

    void write(const Container& cont)
    {
        ContainerWrapper wrapper(cont);
        std::memmove(location_, &wrapper, sizeof(wrapper));
    }
};

}  // namespace bootloader_interface
